import ast

from symdexer.modules import Module


__all__ = "SYM_TYPES", "iter_symbols"


def import_map(node: ast.AST):
    if isinstance(node, (ast.Import, ast.ImportFrom)):
        for alias in node.names:
            yield alias.name


def define_map(node: ast.AST):
    if isinstance(node, (ast.FunctionDef, ast.ClassDef)):
        yield node.name


def assign_map(node: ast.AST):
    if isinstance(node, ast.Assign):
        for target in node.targets:
            if isinstance(target, ast.Name):
                yield target.id


SYM_TYPES = {"imports": import_map, "defines": define_map, "assigns": assign_map}


def iter_symbols(module: Module):
    try:
        root = ast.parse(module.path.read_text("utf-8"))
    except SyntaxError:
        return

    for node in ast.iter_child_nodes(root):
        for sym_type, map_func in SYM_TYPES.items():
            for symbol in map_func(node):
                yield symbol, sym_type



import sqlite3
from pathlib import Path
from typing import Generator

from symdexer.symbols import iter_symbols
from symdexer.modules import walk_modules, Module


__all__ = "make_cache", "search_cache"


def make_cache(file: Path, paths: list[Path]):
    with sqlite3.connect(file) as db:
        db.executescript(
            """
            DROP TABLE IF EXISTS Symbol;

            CREATE TABLE Symbol (
                name TEXT NOT NULL,
                type TEXT NOT NULL,
                module TEXT NOT NULL,
                UNIQUE (name, type, module)
            );
            """
        )

        for path in paths:
            for module in walk_modules(path):
                cache_module(db, module)

        db.commit()


def cache_module(db: sqlite3.Connection, module: Module):
    for symbol, sym_type in iter_symbols(module):
        db.execute(
            "INSERT OR IGNORE INTO Symbol (name, type, module) VALUES (?, ? ,?)",
            (symbol, sym_type, module.name),
        )


def search_cache(file: Path, symbols: list[str], order: list[str]) -> Generator[tuple[str, str], None, None]:
    with sqlite3.connect(file) as db:
        orders = " OR ".join("type = ?" for _ in order)
        patterns = " OR ".join("name LIKE ?" for _ in symbols)
        cursor = db.execute(
            f"""
            SELECT GROUP_CONCAT(name, ", ") as names, module
            FROM Symbol
            WHERE ({patterns}) AND ({orders})
            GROUP BY
                module
            ORDER BY
                LENGTH(module) + LENGTH(names) ASC
            """,
            (*symbols, *order),
        )
        return cursor.fetchall()



from argparse import ArgumentParser

from symdexer.symbols import SYM_TYPES
from symdexer.settings import load_settings
from symdexer.cache import make_cache, search_cache


def find_command(settings_p: str, symbols: list[str], order: list[str]):
    settings = load_settings(settings_p)

    if not settings.cache.exists():
        make_cache(settings.cache, settings.packages)

    for name, module in search_cache(settings.cache, symbols, order or settings.order):
        print(f"from {module} import {name}")


def reset_cache_command(settings_p: str):
    settings = load_settings(settings_p)

    make_cache(settings.cache, settings.packages)

    print("reseted cached")


def init_settings_command(settings_p: str):
    with open(settings_p, "w", encoding="utf-8") as fp:
        fp.write('cache = "symdex.db"\n')
        fp.write('default_order = ["defines", "assigns", "imports"]\n')
        fp.write("\n")
        fp.write("[packages]\n")
        fp.write('package_name = "path/to/package"\n')

    print(f"initialized {settings_p}")


def main():
    parser = ArgumentParser("symdexer")

    parser.add_argument(
        "-s",
        "--settings",
        dest="settings",
        metavar="file",
        default="symdexer.toml",
        help="Name of the file to load the settings from",
    )

    subp = parser.add_subparsers(required=True, dest="command")

    find = subp.add_parser("find", help="Searches the cache for symbols matching a list of patterns")

    find.add_argument(
        "patterns",
        metavar="PATTERN",
        nargs="+",
        help="The patterns to look for"
    )

    find.add_argument(
        "-o",
        "--order",
        metavar="SYM_TYPE",
        choices=list(SYM_TYPES),
        nargs="+"
    )

    subp.add_parser("reset-cache", help="Determines if the cache should be reset or not")

    subp.add_parser("init", help="Creates a new configuration file")

    args = parser.parse_args()

    if args.command == "find":
        find_command(args.settings, args.patterns, args.order)
    elif args.command == "reset-cache":
        reset_cache_command(args.settings)
    elif args.command == "init":
        init_settings_command(args.settings)


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import NamedTuple


__all__ = "Module", "walk_modules"


class Module(NamedTuple):
    path: Path
    name: str


def walk_modules(path: Path, prefix: str = ""):
    if not prefix:
        prefix = path.stem

    if path.is_file():
        if path.suffix == ".py":
            yield Module(path, prefix)
        return

    for subpath in path.iterdir():
        yield from walk_modules(
            subpath,
            prefix if subpath.stem == "__init__" else f"{prefix}.{subpath.stem}",
        )



import tomllib
import importlib
from pathlib import Path
from dataclasses import dataclass

from symdexer.symbols import SYM_TYPES


__all__ = "Settings", "load_settings"


@dataclass
class Settings:
    cache: Path
    order: list[str]
    packages: list[Path]


def load_settings(file: str):
    path = Path(file)

    if not path.is_file():
        raise FileNotFoundError(f"settings file `{file}` does not exist")

    with path.open("rb") as fp:
        data = tomllib.load(fp)

        if any(sym_type not in SYM_TYPES for sym_type in data["default_order"]):
            raise ValueError(f"invalid symbol type in order, values must be {list(SYM_TYPES)}")

        return Settings(Path(data["cache"]), data["default_order"], list(load_packages(data["packages"])))


def load_packages(packages: dict[str, str]):
    for name, path in packages.items():
        if path:
            p = Path(path)

            if not p.is_dir():
                raise ImportError(f"package `{name}` must be a folder")

            yield p
            continue

        p = Path(importlib.import_module(name).__file__)

        if not p.is_file() or p.name.lower() != "__init__.py":
            raise ImportError(f"module `{name}` is not a package")

        yield p.parent
